package test;

import java.io.IOException;
import java.util.Random;
import org.infinispan.Cache;
import org.infinispan.manager.DefaultCacheManager;
import org.infinispan.transaction.tm.DummyTransactionManager;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.BeforeClass;

public class TestSaveMultipleKeys {

    @BeforeClass
    public static void init() {
        System.setProperty("java.net.preferIPv4Stack", "true");
    }
    Cache[] caches;
    DefaultCacheManager[] cacheManagers;
    DummyTransactionManager transactionManager = new DummyTransactionManager();
    int count = 50;
    int threadNumbers = 10;
    int nodes = 3;

    @Before
    public void createCache() throws InterruptedException, IOException {
        cacheManagers = new DefaultCacheManager[nodes];
        caches = new Cache[nodes * 2];

        for (int i = 0; i < nodes; i++) {
            cacheManagers[i] = new org.infinispan.manager.DefaultCacheManager("test-cluster_1.xml");
            Thread.sleep(2000);
            caches[i * 2] = cacheManagers[i].getCache("TEST_CACHE");
            caches[i * 2 + 1] = cacheManagers[i].getCache("TEST_CACHE2");
        }

        System.out.println("===Waiting for clustering join");
        Thread.sleep(2000);
        System.out.println("Start test");
    }

    @After
    public void shutdown() throws Throwable{
        Thread.sleep(10000);
        System.out.println("All test are finished, let's check the locks");
        lockMultipleKeys(0);
        lockMultipleKeys(1);
        lockMultipleKeys(2);

        System.out.println("Stopping all cache managers");
        for (int i = 0; i < cacheManagers.length; i++) {
            cacheManagers[i].stop();
        }
    }

    @Test
    public void testMultipleThreads() throws InterruptedException {

        System.out.println("Start testing....");
        XThread[] threads = new XThread[threadNumbers];
        for (int i = 0; i < threads.length; i++) {
            threads[i] = new XThread(count);
            threads[i].start();
        }

        for (int i = 0; i < threads.length; i++) {
            threads[i].join();
        }
        Thread.sleep(1000);
        if (fail) {
            Assert.fail("failed");
        }
    }

    class XThread extends Thread {

        private int count;

        public XThread(int count) {
            this.count = count;
        }

        @Override
        public void run() {
                for (int i = 0; i < count; i++) {
                    try {
                
                        lockMultipleKeys(random.nextInt(3));
                    } catch (Exception e) {
                        System.out.println("ERROR in loop " + i
                                + ": " + e.getMessage());

                    }
                }
        }
    }

    private void lockMultipleKeys(int i) throws Exception {


        transactionManager.begin();
        try {

            int n = random.nextInt(nodes);
            caches[2 * n].getAdvancedCache().lock(i);
            Integer value = (Integer) caches[2 * n].get(i);
            if (value == null) {
                caches[2 * n].putAsync(i, i);
            } else {
                caches[2 * n].replaceAsync(i, i + 1);
            }

            n = random.nextInt(5);
            if (n < 3) {
                caches[2 * n + 1].getAdvancedCache().lock(i);
                value = (Integer) caches[2 * n + 1].get(i);
                if (value == null) {
                    caches[2 * n + 1].putAsync(i, i);
                } else {
                    caches[2 * n + 1].putAsync(i, i + 1);
                }
            }

            transactionManager.commit();

        } catch (Exception e) {
            fail = true;
            transactionManager.rollback();
            throw e;
        }

    }
    volatile boolean fail = false;
    private Random random = new Random();
}
